# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_magic_cell.ipynb.

# %% auto 0
__all__ = ['system_prompt', 'config', 'chat', 'display_data_to_messages', 'parse_cell_line', 'parse_above_cell',
           'cells_to_messages', 'merge_messages', 'make_cell_from_captured', 'strip_comments', 'handle_config',
           'CellChat', 'fr_cell']

# %% ../nbs/01_magic_cell.ipynb 6
system_prompt="""
You are Claude, a very knowledgeable and intelligent research assistant.
Fulfill simple requests right away.
For more complicated tasks, use <thought> to plan your actions.

When appropriate, use Jupyter notebook. It has python 3.10+ and many packages installed.
Use <code> to add a new code cell at the end of your message. You will receive all its output on the next turn.

Prefer short cells that can be tested quickly.

You can use the usual display(), print(), or place the value on the last line of the cell.
Make sure to not add anything after that value, as it won't be returned if it's not on the last line.
Only output a result once with one method.
Avid using .show() and close matplot figures before the end of the cell - use an explicit display().

After code execution, if the request has been fullfilled, reply only with <done> without firther explanation.
The user will ask follow-up questions if needed. Only do only what they asked for.
You should also end your text with <done> if there is no need to run code.

To show an example in python without running the code, use
```python
python code here
```

For example:
User: Calculate square root of pi
Assistant: <code>
import math
math.sqrt(math.pi)
</code>
User: <stdio>1.7724538509055159<stdio>
Assistant: <done>

User: Display cat.jpg and describe wat you see.
Assistant: <thought>I can use the PIL library for this.</thought><code>
from PIL import Image
Image.open("cat.jpg")
</code>
User: <PIL.PngImagePlugin.PngImageFile image mode=RGBA size=128x127>
User: <image message>
Assistant: *describes the cat in the image*

When writing code, use advanced python features. Assume all packages are installed.
Import modules before using them.
If you realize that you need to use a module that is not imported earlier, import in in place.
"""


# %% ../nbs/01_magic_cell.ipynb 9
import sys
import base64
import json
from typing import List, Dict, Tuple
import textwrap
import warnings

from friendlly.utils.notebook import (
    nbclassic_patch_kernel,
    nbclassic_add_cell,
    nbclassic_update_cell,
    nbclassic_render_cell,
    nbclassic_execute_cell,
    nbclassic_delete_cell,
    nbclassic_get_index,
    update_code_self,
    update_code_next,
    detect_environment,
    vscode_get_cells
    )


from .utils.misc import nict, safe_int
from .utils.capture_trace import strip_junk, SaveTraceback
from .utils.capture_display import PassthroughCapturer, svg_to_png
from .utils.capture_io import TeeIO
import claudette
from claudette import text_msg, img_msg, mk_msg


from IPython import get_ipython
from IPython.display import display, clear_output, Markdown, Javascript

# %% ../nbs/01_magic_cell.ipynb 10
# Patch the Jupyter nbclassic kernel.
if detect_environment() == "nbclassic":
    nbclassic_patch_kernel()


# %% ../nbs/01_magic_cell.ipynb 11
def display_data_to_messages(dsp):
    plain_text = dsp.get("text/plain")
    text = list(filter(lambda x:x[1], [
        ("text-html", dsp.get("text/html")),
        ("latex", dsp.get("text/latex")),
        ("markdown", dsp.get("text/markdown")),
        ("json", dsp.get("application/json")),
        ("js", dsp.get("application/javascript"))
    ]))

    images = list(filter(lambda x:x, [ dsp.get("image/png"), dsp.get("image/jpeg") ]))
    svg = dsp.get("image/svg+xml")
    if svg: images.append(svg_to_png(svg, 256))

    messages = [f"<{t[0]}>{t[1]}</{t[0]}>" for t in text]

    if not len(messages):
        messages = [plain_text]

    claudette_msgs = [text_msg(m) for m in messages]
    if images:
        image = images[0] # Grab only one image
        if isinstance(image, str):
            image = base64.b64decode(image)
        claudette_msgs.append(img_msg(image))

    return claudette_msgs

def parse_cell_line(line):
    num_cells, is_friendlly, is_reply, is_config = 0, False, False, False
    if line.lstrip('# \t').rstrip().startswith("%%fr"):
        is_friendlly = True
        args = line.split()
        for arg in args[1:]:
            if "+" in arg: num_cells = safe_int(arg[1:])
            if arg == "reply": is_reply = True
            if arg == "config" : is_config = True

    return num_cells, is_friendlly, is_reply, is_config

def parse_above_cell(cell):
    assert "source" in cell
    if cell.cell_type != "code":
        assert not "outputs" in cell or not len(cell.outputs), "Got a non-code cell with outputs??"
        return False, False, text_msg(cell.source), []

    num_cells = 0
    is_reply = False
    is_friendlly = False
    code_blk = None
    output_blks = []

    lines = cell.source.strip().splitlines()
    if lines:
        num_cells, is_friendlly, is_reply, _ = parse_cell_line(lines[0])

        # Remove the cell magic
        if is_friendlly:
            lines = lines[1:]
        code_text = "\n".join(lines)

        outputs = cell.get("outputs", [])
        for o in outputs:
            if o.output_type == "stream":
                output_blks.append(text_msg(f"<{o.name}>{o.text}</{o.name}>"))
            if o.output_type in ("display_data", "execute_result"):
                # We produce a lot of empty JS outputs when running the chat, and they don't get
                # cleaned up in time. Let's just ignore empty js objects.
                if "application/javascript" not in o.data or o.data["application/javascript"]:
                    output_blks += display_data_to_messages(o.data)
            if o.output_type == "error":
                output_blks.append(text_msg("<error>" + strip_junk("\n".join(o.traceback)) + "</error>"))



        return is_friendlly, is_reply, code_text, output_blks
    return False, False, None, []

# %% ../nbs/01_magic_cell.ipynb 12
def cells_to_messages(cells):
    messages = []


    # We found the first %%fr message (request from user).
    # This means that we treat further cells (markdown and code without %%fr) as assistant-generated.
    user_message_found = False
    for cell in cells:
        cell_friendlly, cell_reply, code_text, output_blks = parse_above_cell(cell)

        if code_text:
            # If the cell is marked with %%fr reply, or we found the user request above.
            if cell_reply:       role="assistant"  # Messages has '%%fr reply'
            elif cell_friendlly: role="user" # Message has %%fr without 'reply'
            else:                role="assistant" if user_message_found else "user" # It's not an %%fr cell.

            # We found an %%fr cell that is not a reply - means it's a user message!
            if cell_friendlly and not cell_reply: user_message_found=True

            # Assistant code gets wrapped into tags, user code gets wrapped into ```
            # Markdown is passed as is.
            if role == "assistant" and cell.cell_type == "code":
                code_text = f"<code>{code_text}</code>"
            elif role == "user":
                code_text = f"```\n{code_text}\n```"

            messages.append(mk_msg([code_text], role=role))

        # Outputs for both user and assistant code cells are always user messages.
        if output_blks:
            messages.append(mk_msg(output_blks, role="user"))

    return [nict(m) for m in messages]

# %% ../nbs/01_magic_cell.ipynb 13
def merge_messages(msgs):

    # first, merge by role
    role_messages = []
    prev_role = None

    for m in msgs:
        if prev_role == m.role: role_messages[-1].content += m.content
        else:                   role_messages.append(m)
        prev_role = m.role

    # Then, for assistant, also coalesce text messages together
    for m in role_messages:
        merged_content = []
        for c in m.content:
            if c.type == "text" and merged_content and merged_content[-1].type == "text":
                merged_content[-1].text += ("\n" + c.text)
            else: merged_content.append(c)
        m.content = merged_content
    return role_messages


# %% ../nbs/01_magic_cell.ipynb 14
def make_cell_from_captured(source:str, res, stdout:str, stderr:str, displays, tracebacks):
    ip = get_ipython()
    cell = nict()
    cell.metadata  = {}
    cell.cell_type = "code"
    cell.source    = source
    cell.outputs   = []
    if stdout: cell.outputs.append(nict(output_type="stream", name="stdout", text=stdout))
    if stderr: cell.outputs.append(nict(output_type="stream", name="stderr", text=stderr))
    for d in displays:
        cell.outputs.append(nict(output_type="display_data", data=d))
    if res.result:
        formatted, _ = ip.display_formatter.format(res.result)
        cell.outputs.append(nict(output_type="execute_result", data=formatted))
    for t in tracebacks:
        cell.outputs.append(nict(t, output_type="error"))

    return cell

# %% ../nbs/01_magic_cell.ipynb 15
config = nict(
    environment=detect_environment(),
    autorun=detect_environment() == "nbclassic",
    md_cells=detect_environment() == "nbclassic",
    comment_after_run=True,
    api_key=None,
    model="claude-3-5-sonnet-20240620",
    system_prompt=system_prompt,
)

def strip_comments(cell: str):
    lines = cell.splitlines()
    return "\n".join([ l.split("#")[0] for l in lines ]).strip()

def handle_config(cell):
    global config
    ip = get_ipython()

    cell = strip_comments(cell)
    # If the cell was empty, show the config
    if not cell:
        cell = "\n".join([f'{k}="""{v}"""' if isinstance(v, str) else f"{k}={v}" for k, v in config.items()])
        update_code_self("%%fr config\n" + cell)
    else:
        new_config = {}
        # Code, globals, locals. We want the locals
        exec(cell, {}, new_config)
        config.update(new_config)

        if config.autorun and config.environment != "nbclassic":
            warnings.warn("Autorun is only supported in nbclassic environment at the moment.")

        if config.md_cells and config.environment != "nbclassic":
            warnings.warn("Inserting Markdown cells is only supported in nbclassic environment at the moment.")
        if config.api_key:
            warnings.warn("FIY It's safer to set ANTHROPIC_API_KEY in the environment or even better .env file and use dotenv package to load it.")

# %% ../nbs/01_magic_cell.ipynb 16
class CellChat():
    def __init__(self, config):
        self.config = config
        self.display_handle = None # Used when config.md_cell=False
        self.history = []
        self.client = claudette.Client(model=self.config.model)

    def update_markdown(self, text):
        # md_cells means we output the text as markdown cells.
        # Otherwise, we display them.
        if self.config.md_cells and self.idx is not None:
            if not self.added_md_cell:
                nbclassic_add_cell(self.idx+1, "markdown")
                self.added_md_cell = True
            nbclassic_update_cell(self.idx+1, text)
            nbclassic_render_cell(self.idx+1)
        else:
            data = {"text/markdown": text, "config/message_type": "assistant"}
            if self.display_handle:
                self.display_handle.update(data, raw=True)
            else:
                self.display_handle = display(data, raw=True, display_id=True)

    def delete_markdown(self):
        if self.config.md_cells and self.added_md_cell:
            nbclassic_delete_cell(self.idx+1)
            self.added_md_cell = False
        elif self.display_handle:
            # Empty JS cells ge cleaned up after the cell is executed.
            self.display_handle.update(Javascript(""))

    def update_code(self, code):
        # We have an markdown cell before the code cell. This can only be done with JS in nbclassic.
        if self.added_md_cell and self.idx is not None:
            if not self.added_code_cell:
                nbclassic_add_cell(self.idx+2, "code")
                self.added_code_cell = True

            nbclassic_update_cell(self.idx+2, code)
        else:
            # No markdown cell before core - we can use the native ipython method.
            update_code_next(code)

    def go(self, idx,  num_above, messages):
        self.idx = idx
        self.added_md_cell = self.added_code_cell = False
        self.thought_display_id = None
        self.history.append(messages)

        tokens = []
        try:
            md_reply = ""
            md_updated = False
            code_reply = ""
            code_updated = False
            thought = ""

            state = "md"

            r = self.client(messages,
                            sp=self.config.system_prompt,
                            stream=True,
                            stop=["</code>", "<done>"])
            for token in r:
                tokens.append(token)
#                 print(state, f"'{token}'", f"'{code_reply}'", f"'{md_reply}'")

                if state == "md":
                    md_reply += token
                    md_updated = True

                    if "<code>" in md_reply:
                        md_reply, code_reply = md_reply.split("<code>")
                        code_updated = True
                        state = "code"

                    if "<thought>" in md_reply:
                        md_reply_body, thought = md_reply.split("<thought>")
                        state = "thought"

                elif state == "thought":
                    thought += token
                    md_tail = ""
                    md_updated = True

                    def combine_md_reply(body, thought, tail=""):
                        return f'{body}<details><summary style="display: list-item">🧠</summary>{thought}</details>{tail}'

                    if "</thought>" in thought:
                        thought, md_tail = thought.split("</thought>")
                        state = "md"
                    md_reply = combine_md_reply(md_reply_body, thought, md_tail)

                else: #code
                    code_reply += token
                    code_updated = True

                if md_updated:
                    md_updated = False
                    if md_reply.strip():
                        self.update_markdown(md_reply.strip())
                    else:
                        self.delete_markdown()

                if code_updated:
                    code_updated = False
                    prefix = f"%%fr +{num_above + 2 if self.added_md_cell else 1} reply\n"
                    self.update_code(prefix + code_reply.strip())

            if self.client.stop_reason == "stop_sequence":
                tokens += self.client.stop_sequence
            self.history.append(claudette.mk_msg("".join(tokens), role="assistant"))

            code_cell_idx = None if self.idx is None else self.idx + int(self.added_code_cell) + int(self.added_md_cell)
            return code_cell_idx, md_reply

        except BaseException as e:
            raise e
#             display(Markdown(f"🚫 {repr(e)}"))
#             get_ipython().showtraceback()

# %% ../nbs/01_magic_cell.ipynb 17
chat = None
def fr_cell(line=None, cell=None):
    global chat
    global config
    chat = CellChat(config=config)

    num_cells, is_friendlly, is_reply, is_config = 0, False, False, False
    idx = None
    ip = get_ipython()
    prev_messages = []
    if line:
        # line comes with magic already stripped.
        num_cells, is_friendlly, is_reply, is_config = parse_cell_line("%%fr " + line)
        if num_cells and not is_config:
            assert is_friendlly, "Called on a cell without %%fr?"
            header = nict(ip.parent_header)


            cells_above = []
            if config.environment == "nbclassic":
                if "cells_above" not in header.content:
                    warnings.warn("Jupyter did not send any of the above cells")
                else:
                    cells_above = header.content.cells_above;
            elif config.environment == "vscode":
                idx, cells_above = vscode_get_cells(num_cells)
            else:
                warnings.warn("Only nbclassic and vscode environments can access previous cells")
            prev_messages = cells_to_messages(cells_above)

    if is_config:
        handle_config(cell)
        return

    if idx is None:
        if config.environment == "nbclassic":
            idx = nbclassic_get_index()
        elif config.environment == "vscode":
            idx, _ = vscode_get_cells(0)

    # If it's a folllow-up message, we should first run the current cell and collect its outputs,
    # then pass the outputs to the chat as user messages for further generation.

    if is_reply:
        disp_capturer = PassthroughCapturer()
        disp_capturer.start()

        sys.stdout = stdout = TeeIO(sys.stdout)
        sys.stderr = stderr = TeeIO(sys.stderr)

        traceback = SaveTraceback(ip._showtraceback)
        traceback.register()

        # This runs the current cell, and we should capture all its outputs.
        res = ip.run_cell(cell)

        traceback.unregister()
        disp_capturer.stop()
        sys.stdout = stdout._stream
        sys.stderr = stderr._stream

        new_cell = make_cell_from_captured(source="%%fr " + line + "\n" + cell,
                                           res=res,
                                           stdout=stdout.getvalue(),
                                           stderr=stderr.getvalue(),
                                           displays=disp_capturer.get_outputs(),
                                           tracebacks=traceback.tracebacks)
        prev_messages += cells_to_messages([new_cell])

    else:
        prev_messages.append(nict(mk_msg(cell, role="user")))

    prev_messages = merge_messages(prev_messages)
    # Assistant message can't be the first in the conversation, nor can it be the last.
    if prev_messages:
        if prev_messages[0].role == "assistant":
            prev_messages.insert(0, nict(mk_msg("(Beginning of the conversation truncated)", role="user")))
        if prev_messages[-1].role == "assistant":
            prev_messages.append(nict(mk_msg("Continue", role="user")))

    code_idx, md_text = chat.go(idx, num_cells, prev_messages)

    # If it's a reply cell, comment out the magic. Otherwise, comment out the
    # whole cell, as it's likely plain text, and not valid Python.
    # This makes it easy to re-run the ai-generated notebook manually without triggering the ai.
    cell_text = f"%%fr {line}\n{cell.strip()}".splitlines()
    if is_reply:
        cell_text[0] = "# " + cell_text[0]
    elif config.comment_after_run:
        cell_text = ["# " + l for l in cell_text]

    update_code_self("\n".join(cell_text))

    # This schedules the execution of the next cell.
    if config.autorun and idx is not None and code_idx != idx:
        nbclassic_execute_cell(code_idx)


    # We've created a lot of invisible JS display outputs. Unfortunaely I don't think there is a way
    # to clear them without removing everying.
    clear_output()

    if is_reply:
        # Re-display the outputs.
        for d in disp_capturer.raw_outputs:
            ip.display_pub.publish(**d)
        if stdout.getvalue(): print(stdout.getvalue())
        if stderr.getvalue(): print(stderr.getvalue(), file=sys.stderr)
        for tb in traceback.tracebacks:
            ip._showtraceback(tb.ename, tb.evalue, tb.traceback)

    if md_text: display(Markdown(md_text))
    if is_reply: return res.result




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/04_utils.cell_cbs.ipynb.

# %% auto 0
__all__ = ['ExecCBs']

# %% ../../nbs/04_utils.cell_cbs.ipynb 2
import sys
from IPython import get_ipython
from . import PassthroughCapturer, TeeIO, SaveTraceback

# %% ../../nbs/04_utils.cell_cbs.ipynb 4
class ExecCBs:
    stdout = None
    stderr = None
    result = None # The thing at the end of the cell
    display =None
    saved_streams = None

    def __init__(self, after_run=None, once=True):
        self.do_after_run = after_run
        self.once = once

    def pre_callback(self, info=None):
        # print("# pre")
        ip = get_ipython()
        # This will capture diplay() and friends
        self.display = PassthroughCapturer()
        self.display.start()

        # This will capture stdout/stderr
        sys.stdout = self.stdout = TeeIO(sys.stdout)
        sys.stderr = self.stderr = TeeIO(sys.stderr)


        self.tracebacks = SaveTraceback(ip._showtraceback)
        self.tracebacks.register()

        # We register the callback here, so it does not run after the cell
        # that registered the pre

        ip.events.register("post_run_cell", self.post_callback)
        # We get the result in the post callback info, no need to register anything else.

    def post_callback(self, result):
        # Post will be called in the cell that registers the
        # callbacks, but it wuld not have ran pre.
        # print("# post")
        self.result = result

        if self.stdout:
            sys.stdout = self.stdout._stream
            sys.stderr = self.stderr._stream

        if self.display:
            self.display.stop()

        if self.tracebacks:
            self.tracebacks.unregister()

        if self.once:
            self.unregister()

        if self.do_after_run:
            self.do_after_run(
                self.stdout.getvalue(),
                self.stderr.getvalue(),
                self.display.get_outputs(),
                result,
                self.tracebacks.tracebacks
            )

    def register(self):
        ip = get_ipython()
        ip.events.register("pre_run_cell", self.pre_callback)


    def unregister(self):
        try:
            ip = get_ipython()
            ip.events.unregister("pre_run_cell", self.pre_callback)
            ip.events.unregister("post_run_cell", self.post_callback)
        except: pass # In case unregister was aclled before a cell ran, and we never regisstered pre.




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03_utils.capture_trace.ipynb.

# %% auto 0
__all__ = ['strip_junk', 'SaveTraceback']

# %% ../../nbs/03_utils.capture_trace.ipynb 2
import re
from IPython import get_ipython
from .misc import nict

# %% ../../nbs/03_utils.capture_trace.ipynb 3
def strip_junk(text):
    # This line just easts up tokens. We don't need it.
    text = text.replace("---------------------------------------------------------------------------", "")
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)


# %% ../../nbs/03_utils.capture_trace.ipynb 5
class SaveTraceback:
    def __init__(self, saved):
        self._saved_showtraceback = saved
        self.tracebacks = []

    def _showtraceback(self, etype, value, stb):
        self.tracebacks.append( nict(
            ename=etype,
            evalue=", ".join(value.args),
            traceback= stb # ["\n".join([ strip_junk(s) for s in stb]) if isinstance(stb, list) else stb]
        ))
        self._saved_showtraceback(etype, value, stb)

    def register(self):
        get_ipython()._showtraceback = self._showtraceback

    def unregister(self):
        get_ipython()._showtraceback = self._saved_showtraceback




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/05_utils.notebook.ipynb.

# %% auto 0
__all__ = ['extract_vscode_path', 'vscode_get_cells', 'update_code_self', 'update_code_next', 'detect_environment',
           'nbclassic_patch_kernel', 'nbclassic_add_cell', 'nbclassic_update_cell', 'nbclassic_execute_cell',
           'nbclassic_render_cell', 'nbclassic_delete_cell', 'nbclassic_get_index']

# %% ../../nbs/05_utils.notebook.ipynb 2
from IPython import get_ipython
from IPython.display import display, clear_output, Markdown, Javascript
import json
import os
import time
import psutil
from urllib.parse import urlparse
import warnings
from .misc import nict

# %% ../../nbs/05_utils.notebook.ipynb 3
def extract_vscode_path():
    """
    Extracts the filename from the parent_header of the current notebook.
    """
    cellid = get_ipython().parent_header.get("metadata", {}).get("cellId", '')
    url = urlparse(cellid)
    return url.path


# %% ../../nbs/05_utils.notebook.ipynb 4
def vscode_get_cells(num_cells):
    path = extract_vscode_path()

    def get_last_modified(path):
        try: return os.path.getmtime(path)
        except BaseException as e:
            warnings.warn(f"{e}: Could not get mtime for {path}")
            return 0

    if path:
        ip = get_ipython()
        display(Javascript("")) # Empty js to kick off autosave.

        last_modified = get_last_modified(path)
        current_timestamp = time.time()
        assert current_timestamp - last_modified > 0, f"timestamp on file {path} is in the future"


        # If the file has been modified more than a second ago, wait a bit for it to be saved, but not too long
        if current_timestamp - last_modified > 1:
            for i in range(50):
                time.sleep(0.1)
                last_modified = get_last_modified(path)
                if current_timestamp - last_modified < 1:
                    break
            else:
                warnings.warn("Make sure autosave is set to afterDelay in vscode settings, and the delay is less than a second!")

        for i in range(50):
            try: # In case the file is being written to and is not parseable.
                with open(path) as f:
                    data = json.load(f)
                    cells = data.get("cells", [])
                    for idx, cell in enumerate(cells):
                        if cell.get("execution_count") == ip.execution_count:
                            cells = [ nict(c) for c in cells[(max(0, idx-num_cells)):idx] ]
                            for c in cells: c.source = "\n".join(c.source) # source is saved as an array of strings in jupyter.
                            return idx, cells
            except: pass
            time.sleep(0.1)

        warnings.warn(f"Could not find the cell data in {path} . The assistant won't be able to see previous cells.")
    else:
        warnings.warn("VSCode did not send a cellId in the parent_header. The assistant won't be able to see previous cells.")

    return None, []

# %% ../../nbs/05_utils.notebook.ipynb 5
def update_code_self(source:str):
    """
    Updates the current cell with the contents of the cell passed as argument.
    """

    payload = dict(
        source="set_next_input",
        text=source,
        replace=True,
        clear_output=False
    )
    # We have to call payload_manager directly because ip.set_next_input does not have a clear_output parameter
    get_ipython().payload_manager.write_payload(payload, single=False)


# %% ../../nbs/05_utils.notebook.ipynb 6
def update_code_next(source:str):
    """
    Updates the next cell with the code passed as argument.
    """
    payload = dict(
        source="set_next_input",
        text=source,
        replace=False,
        clear_output=False
    )
    # We have to call payload_manager directly because ip.set_next_input does not have a clear_output parameter
    get_ipython().payload_manager.write_payload(payload, single=True)

# %% ../../nbs/05_utils.notebook.ipynb 7
def detect_environment():
    """
    Detects if we are running in vscode, Jupyter nbclassic, or Jupyter notebook 7 / Jupyter lab.
    """

    parent_cmdline = psutil.Process(os.getppid()).cmdline()

    for l in parent_cmdline:
        # Vscode might have a different name (e,g, cursor), but it should have "vscode" somewhere in the command line.
        if "vscode" in l.lower():
            return "vscode"

        # jupyter-nbclassic is the modern name of the "old" Jupyter notebook, as it was in Jupyter 6.
        if "jupyter-nbclassic" in l:
            return "nbclassic"

        # jupyter-notebook is probably Jupyter notebook 7, which is a re-skin of Jupyter lab
        if "jupyter-lab" in l or "jupyter-notebook" in l:
            return "jupyterlab"

    warnings.warn("Could not detect environment. Assuming Jupyter nbclassic, but some features may not work.")
    return "nbclassic"

# %% ../../nbs/05_utils.notebook.ipynb 8
def inject_js(js:str):
    """Inject some javascript into the notebook and clear the output to prevent it from running on reload"""

    display_handle = display(Javascript(js), display_id=True)
    # If we keep the outpur, it will be run on every notebook load.
    # This leaves empty js outputs that should be cleaned up separately.
    display_handle.update(Javascript(""))

# %% ../../nbs/05_utils.notebook.ipynb 9
def nbclassic_patch_kernel():
    payload = """
    console.log("patching nbclassic execute function...")
    Jupyter.CodeCell.prototype.execute = function (stop_on_error) {
        if (!this.kernel) {
            console.log(i18n.msg._("Can't execute cell since kernel is not set."));
            return;
        }

        if (stop_on_error === undefined) {
            if (this.metadata !== undefined &&
                    this.metadata.tags !== undefined) {
                if (this.metadata.tags.indexOf('raises-exception') !== -1) {
                    stop_on_error = false;
                } else {
                    stop_on_error = true;
                }
            } else {
               stop_on_error = true;
            }
        }

        this.clear_output(false, true);
        var old_msg_id = this.last_msg_id;
        if (old_msg_id) {
            this.kernel.clear_callbacks_for_msg(old_msg_id);
            delete Jupyter.CodeCell.msg_cells[old_msg_id];
            this.last_msg_id = null;
        }
        if (this.get_text().trim().length === 0) {
            // nothing to do
            this.set_input_prompt(null);
            return;
        }
        this.set_input_prompt('*');
        this.element.addClass("running");
        var callbacks = this.get_callbacks();


        const cell_index = Jupyter.notebook.find_cell_index(this)

        let extras = {
            cell_index : cell_index,
            cell_id: this.id
        }
        let text = this.get_text().trim()
        let firstLine = text.split('\\n')[0];

        // Parse the magic command
        if (firstLine.startsWith("%%fr")) {
            // Separate args by spaces or tabs
            let parts = firstLine.split(/\\s+|\\t+/);
            let magic = parts[0];
            if (parts.length > 1) {
                let magic_args = parts.slice(1);
                let plusNArg = magic_args.find(arg => arg.startsWith('+') && !isNaN(parseInt(arg.slice(1))));
                if (plusNArg) {
                    let n = parseInt(plusNArg.slice(1));
                    let start_pos = Math.max(0, cell_index - n);
                    cells = Jupyter.notebook.get_cells().slice(start_pos, cell_index);

                    extras = {
                        cells_above: cells,
                        ...extras
                    }
                }
            }
        }

        this.last_msg_id = this.kernel.execute(
            this.get_text(),
            callbacks,
            {silent: false, store_history: true, stop_on_error : stop_on_error, ...extras });
        Jupyter.CodeCell.msg_cells[this.last_msg_id] = this;
        this.render();
        this.events.trigger('execute.CodeCell', {cell: this});
        var that = this;
        function handleFinished(evt, data) {
            if (that.kernel.id === data.kernel.id && that.last_msg_id === data.msg_id) {
                    that.events.trigger('finished_execute.CodeCell', {cell: that});
                that.events.off('finished_iopub.Kernel', handleFinished);
              }
        }
        this.events.on('finished_iopub.Kernel', handleFinished);
    };
    Jupyter.notebook.events.trigger('set_dirty.Notebook', {value: true});
    console.log("Done.")
    """
    inject_js(payload)

# %% ../../nbs/05_utils.notebook.ipynb 10
def nbclassic_add_cell(
        idx:int = None, # Index of the cell to add. If none, add the cell under the selected one.
        cell_type:str = "code" # Type of cell to add. Can be "code", "markdown", "raw"
    ):
    """
    Add a new notebook cell.
    """
    if not idx:
        index_payload = "let index = Jupyter.notebook.get_selected_index()+1;"
    else:
        index_payload = f"let index = {idx}"

    payload = f"""
    {index_payload}

    Jupyter.notebook.insert_cell_at_index("{cell_type}", index)
    let cell = Jupyter.notebook.get_cell(index);
    cell.events.trigger('set_dirty.Notebook', {{value: true}});
    """

    inject_js(payload)

# %% ../../nbs/05_utils.notebook.ipynb 11
def nbclassic_update_cell(
    idx:int, # Index of the cell to update. None to update the current cell
    text:str, # Text to set in the cell
    flush:bool = True # Notify Jupyter that the cell has been updated.
    ):

    def escape_for_js(text):
        # Use json.dumps to escape the string for JavaScript
        escaped = json.dumps(text)
        # Remove the surrounding quotes added by json.dumps
        escaped = escaped[1:-1]
        # Escape backticks and ${} sequences
        return escaped.replace('`', '\\`').replace('${', '\\${')

    payload = f"""
    let cell = Jupyter.notebook.get_cell({idx})
    cell.set_text(`{escape_for_js(text)}`)
    """
    if flush:
         payload = payload + "\nJupyter.notebook.events.trigger('set_dirty.Notebook', {value: true});"
    inject_js(payload)

# %% ../../nbs/05_utils.notebook.ipynb 12
def nbclassic_execute_cell(
        idx:int # Index of the cell to execute. They start at 0
    ):
    payload = f"""
    console.log("execute_cell", {idx});
    Jupyter.notebook.events.trigger('set_dirty.Notebook', {{value: true}});
    let cell = Jupyter.notebook.get_cell({idx})
    cell.execute()
    """
    # tt = display(f"About to run the cell {idx}...", display_id=True)
    inject_js(payload)

# %% ../../nbs/05_utils.notebook.ipynb 13
def nbclassic_render_cell(idx:int): # Cell to render.
    """Render a markdown cell"""

    payload = f"""
    let cell = Jupyter.notebook.get_cell({idx})
    cell.unrender()
    Jupyter.notebook.events.trigger('set_dirty.Notebook', {{value: true}});

    cell.render()

    """

    inject_js(payload)

# %% ../../nbs/05_utils.notebook.ipynb 14
def nbclassic_delete_cell(idx:int): # Cell to delete.
    """Delete a cell"""

    payload = f"""
    console.log("deleting cell", {idx});
    Jupyter.notebook.delete_cell({idx});
    Jupyter.notebook.events.trigger('set_dirty.Notebook', {{value: true}});
    """

    inject_js(payload)

# %% ../../nbs/05_utils.notebook.ipynb 15
def nbclassic_get_index():
    ip = get_ipython()
    return ip.parent_header["content"].get("cell_index", -1)



__version__ = "0.2.0"

from .magic_cell import fr_cell
from .magic_line import fr_line
from .utils.notebook import nbclassic_patch_kernel, detect_environment
from IPython.display import clear_output

def load_ipython_extension(ipython):
    # if detect_environment() == "nbclassic":
    #     nbclassic_patch_kernel()
    #     clear_output()

    ipython.register_magic_function(fr_cell, "cell", magic_name="fr")
    ipython.register_magic_function(fr_line, "line", magic_name="fr")

def unload_ipython_extension(ipython):
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/02_utils.capture_display.ipynb.

# %% auto 0
__all__ = ['PassthroughCapturer', 'svg_to_png', 'display_to_messages']

# %% ../../nbs/02_utils.capture_display.ipynb 2
from IPython.core.displaypub import DisplayPublisher
from IPython.core.display_functions import _new_id
from IPython.display import display, clear_output, Javascript
from IPython import get_ipython
from typing import List
from PIL import Image
from collections import defaultdict
import warnings
import sys
import re
import cairosvg
import xml.etree.ElementTree as ET
from io import BytesIO
import base64

from .misc import nict

# %% ../../nbs/02_utils.capture_display.ipynb 3
class PassthroughCapturer(DisplayPublisher):
    """A DisplayPublisher that stores AND displays!"""

    raw_outputs: List

    def __init__(self):
        self._publisher = None
        self.raw_outputs = []

    def publish(
        self, data, metadata=None, source=None, *, transient=None, update=False
    ):
        self.raw_outputs.append(nict(data=data, metadata=metadata, transient=transient, update=update))
        self._publisher.publish(data, metadata=metadata, transient=transient, update=update)

    def clear_output(self, wait=False):
        self.raw_outputs = []
        self._publisher.clear_output(wait)

    def start(self):
        ip = get_ipython()
        self._publisher = ip.display_pub
        ip.display_pub = self

        # print("starting passthrough: ", id(self._publisher), id(self))

    # Note: stop must be called from the cell that started the
    # passthrough, or from the associated post-cell callback
    def stop(self):
        ip = get_ipython()
        assert self._publisher
        ip.display_pub = self._publisher

    def get_outputs(self, with_js=False):
        coalesced_outputs = []
        display_id_map = defaultdict(list)

        outputs = self.raw_outputs
        if not with_js:
            outputs = [o for o in outputs if 'application/javascript' not in o['data']]

        # dicts are ordered in Python 3.7+
        for output in outputs:
            display_id = output['transient']['display_id'] if output['transient'] else _new_id()
            if not output['update']:
                display_id_map[display_id] += [output["data"]]
            else:
                display_id_map[display_id][-1] = output["data"]

        for display_id, output_list in display_id_map.items():
            coalesced_outputs += output_list

        return coalesced_outputs

# %% ../../nbs/02_utils.capture_display.ipynb 6
from claudette import text_msg, img_msg

# %% ../../nbs/02_utils.capture_display.ipynb 7
def svg_to_png(svg_content, max_size):
    root = ET.fromstring(svg_content)


    def parse_dimension(value):
        if value.endswith('%'):
            return float(value[:-1]) / 100 * max_size
        return float(re.findall(r'[\d.]+', value)[0])

    width = parse_dimension(root.attrib.get('width', str(max_size)))
    height = parse_dimension(root.attrib.get('height', str(max_size)))

    viewBox = root.attrib.get('viewBox')
    if viewBox:
        _, _, vb_width, vb_height = map(float, viewBox.split())
        if width == max_size:
            width = vb_width
        if height == max_size:
            height = vb_height

    scale = min(max_size / width, max_size / height)
    new_width = int(width * scale)
    new_height = int(height * scale)

    # Render PNG
    png =  cairosvg.svg2png(bytestring=svg_content, output_width=new_width, output_height=new_height)
    return png

# %% ../../nbs/02_utils.capture_display.ipynb 9
def display_to_messages(dsp):
    plain_text = dsp.get("text/plain")
    text = list(filter(lambda x:x[1], [
        ("text-html", dsp.get("text/html")),
        ("latex", dsp.get("text/latex")),
        ("markdown", dsp.get("text/markdown")),
        ("json", dsp.get("application/json")),
        ("js", dsp.get("application/javascript"))
    ]))

    images = list(filter(lambda x:x, [ dsp.get("image/png"), dsp.get("image/jpeg") ]))
    svg = dsp.get("image/svg+xml")
    if svg: images.append(svg_to_png(svg, 256))

    messages = [f"<{t[0]}>{t[1]}</{t[0]}>" for t in text]

    if not len(messages):
        messages = [plain_text]

    claudette_msgs = [text_msg(m) for m in messages]
    if images:
        imagebytes = base64.b64decode(images[0]) if not isinstance(images[0], bytes) else images[0]
        claudette_msgs.append(img_msg(imagebytes)) # Grab only one image

    return claudette_msgs



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/06_utils.misc.ipynb.

# %% auto 0
__all__ = ['nict', 'GenericEncoder', 'safe_int']

# %% ../../nbs/06_utils.misc.ipynb 1
import pprint

# %% ../../nbs/06_utils.misc.ipynb 2
class nict(dict):
    def __init__(self, d=dict(), /, **kwargs):
        if d is None: d = {}
        else:
            assert isinstance(d, dict), f'expected dict, got {type(d)}'
            d = d.copy()

        d.update(kwargs)
        for k,v  in d.items():
            if isinstance(v, dict):
                d[k] = nict(**v)
            elif isinstance(v, list):
                d[k] = [nict(**item) if isinstance(item, dict) else item for item in v]
            elif isinstance(v, tuple):
                d[k] = tuple(nict(**item) if isinstance(item, dict) else item for item in v)
            else:
                d[k] = v

        super().__init__(**d)

    def __getattr__(self, key):
        return self[key] if key in self else super().__getattribute__(key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]

    def __dir__(self):
        return list(self.keys()) + dir(super())

    def _to_dict(self):
        return {k: v._to_dict() if isinstance(v, nict) else v for k, v in self.items()}

# %% ../../nbs/06_utils.misc.ipynb 13
import json
from datetime import datetime

# %% ../../nbs/06_utils.misc.ipynb 14
class GenericEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        try:
            return obj.__dict__
        except AttributeError:
            try:
                return str(obj)
            except:
                return f"<{type(obj).__name__}>"

# %% ../../nbs/06_utils.misc.ipynb 15
def safe_int(s, default=0):
    try:
        return int(s)
    except ValueError:
        return default



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/friendlly',
                'doc_host': 'https://xl0.github.io',
                'git_url': 'https://github.com/xl0/friendlly',
                'lib_path': 'friendlly'},
  'syms': { 'friendlly.display': { 'friendlly.display.PassthroughCapturer': ('display.html#passthroughcapturer', 'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.__init__': ( 'display.html#passthroughcapturer.__init__',
                                                                                       'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.clear_output': ( 'display.html#passthroughcapturer.clear_output',
                                                                                           'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.get_outputs': ( 'display.html#passthroughcapturer.get_outputs',
                                                                                          'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.publish': ( 'display.html#passthroughcapturer.publish',
                                                                                      'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.start': ( 'display.html#passthroughcapturer.start',
                                                                                    'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.stop': ( 'display.html#passthroughcapturer.stop',
                                                                                   'friendlly/display.py')},
            'friendlly.magic_cell': { 'friendlly.magic_cell.CellChat': ('magic_cell.html#cellchat', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.__init__': ( 'magic_cell.html#cellchat.__init__',
                                                                                  'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.delete_markdown': ( 'magic_cell.html#cellchat.delete_markdown',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.go': ('magic_cell.html#cellchat.go', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.update_code': ( 'magic_cell.html#cellchat.update_code',
                                                                                     'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.update_markdown': ( 'magic_cell.html#cellchat.update_markdown',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.cells_to_messages': ( 'magic_cell.html#cells_to_messages',
                                                                                  'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.display_data_to_messages': ( 'magic_cell.html#display_data_to_messages',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.fr_cell': ('magic_cell.html#fr_cell', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.handle_config': ('magic_cell.html#handle_config', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.make_cell_from_captured': ( 'magic_cell.html#make_cell_from_captured',
                                                                                        'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.merge_messages': ('magic_cell.html#merge_messages', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.parse_above_cell': ( 'magic_cell.html#parse_above_cell',
                                                                                 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.parse_cell_line': ( 'magic_cell.html#parse_cell_line',
                                                                                'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.strip_comments': ('magic_cell.html#strip_comments', 'friendlly/magic_cell.py')},
            'friendlly.magic_line': { 'friendlly.magic_line.fr_line': ('magic_line.html#fr_line', 'friendlly/magic_line.py'),
                                      'friendlly.magic_line.parse_cell': ('magic_line.html#parse_cell', 'friendlly/magic_line.py')},
            'friendlly.utils.capture_display': { 'friendlly.utils.capture_display.PassthroughCapturer': ( 'utils.capture_display.html#passthroughcapturer',
                                                                                                          'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.PassthroughCapturer.__init__': ( 'utils.capture_display.html#passthroughcapturer.__init__',
                                                                                                                   'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.PassthroughCapturer.clear_output': ( 'utils.capture_display.html#passthroughcapturer.clear_output',
                                                                                                                       'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.PassthroughCapturer.get_outputs': ( 'utils.capture_display.html#passthroughcapturer.get_outputs',
                                                                                                                      'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.PassthroughCapturer.publish': ( 'utils.capture_display.html#passthroughcapturer.publish',
                                                                                                                  'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.PassthroughCapturer.start': ( 'utils.capture_display.html#passthroughcapturer.start',
                                                                                                                'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.PassthroughCapturer.stop': ( 'utils.capture_display.html#passthroughcapturer.stop',
                                                                                                               'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.display_to_messages': ( 'utils.capture_display.html#display_to_messages',
                                                                                                          'friendlly/utils/capture_display.py'),
                                                 'friendlly.utils.capture_display.svg_to_png': ( 'utils.capture_display.html#svg_to_png',
                                                                                                 'friendlly/utils/capture_display.py')},
            'friendlly.utils.capture_io': { 'friendlly.utils.capture_io.TeeIO': ( 'utils.capture_io.html#teeio',
                                                                                  'friendlly/utils/capture_io.py'),
                                            'friendlly.utils.capture_io.TeeIO.__getattr__': ( 'utils.capture_io.html#teeio.__getattr__',
                                                                                              'friendlly/utils/capture_io.py'),
                                            'friendlly.utils.capture_io.TeeIO.__init__': ( 'utils.capture_io.html#teeio.__init__',
                                                                                           'friendlly/utils/capture_io.py'),
                                            'friendlly.utils.capture_io.TeeIO.__setattr__': ( 'utils.capture_io.html#teeio.__setattr__',
                                                                                              'friendlly/utils/capture_io.py'),
                                            'friendlly.utils.capture_io.TeeIO.getvalue': ( 'utils.capture_io.html#teeio.getvalue',
                                                                                           'friendlly/utils/capture_io.py'),
                                            'friendlly.utils.capture_io.TeeIO.write': ( 'utils.capture_io.html#teeio.write',
                                                                                        'friendlly/utils/capture_io.py')},
            'friendlly.utils.capture_trace': { 'friendlly.utils.capture_trace.SaveTraceback': ( 'utils.capture_trace.html#savetraceback',
                                                                                                'friendlly/utils/capture_trace.py'),
                                               'friendlly.utils.capture_trace.SaveTraceback.__init__': ( 'utils.capture_trace.html#savetraceback.__init__',
                                                                                                         'friendlly/utils/capture_trace.py'),
                                               'friendlly.utils.capture_trace.SaveTraceback._showtraceback': ( 'utils.capture_trace.html#savetraceback._showtraceback',
                                                                                                               'friendlly/utils/capture_trace.py'),
                                               'friendlly.utils.capture_trace.SaveTraceback.register': ( 'utils.capture_trace.html#savetraceback.register',
                                                                                                         'friendlly/utils/capture_trace.py'),
                                               'friendlly.utils.capture_trace.SaveTraceback.unregister': ( 'utils.capture_trace.html#savetraceback.unregister',
                                                                                                           'friendlly/utils/capture_trace.py'),
                                               'friendlly.utils.capture_trace.strip_junk': ( 'utils.capture_trace.html#strip_junk',
                                                                                             'friendlly/utils/capture_trace.py')},
            'friendlly.utils.cell_cbs': { 'friendlly.utils.cell_cbs.ExecCBs': ( 'utils.cell_cbs.html#execcbs',
                                                                                'friendlly/utils/cell_cbs.py'),
                                          'friendlly.utils.cell_cbs.ExecCBs.__init__': ( 'utils.cell_cbs.html#execcbs.__init__',
                                                                                         'friendlly/utils/cell_cbs.py'),
                                          'friendlly.utils.cell_cbs.ExecCBs.post_callback': ( 'utils.cell_cbs.html#execcbs.post_callback',
                                                                                              'friendlly/utils/cell_cbs.py'),
                                          'friendlly.utils.cell_cbs.ExecCBs.pre_callback': ( 'utils.cell_cbs.html#execcbs.pre_callback',
                                                                                             'friendlly/utils/cell_cbs.py'),
                                          'friendlly.utils.cell_cbs.ExecCBs.register': ( 'utils.cell_cbs.html#execcbs.register',
                                                                                         'friendlly/utils/cell_cbs.py'),
                                          'friendlly.utils.cell_cbs.ExecCBs.unregister': ( 'utils.cell_cbs.html#execcbs.unregister',
                                                                                           'friendlly/utils/cell_cbs.py')},
            'friendlly.utils.misc': { 'friendlly.utils.misc.GenericEncoder': ('utils.misc.html#genericencoder', 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.GenericEncoder.default': ( 'utils.misc.html#genericencoder.default',
                                                                                       'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict': ('utils.misc.html#nict', 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict.__delattr__': ( 'utils.misc.html#nict.__delattr__',
                                                                                 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict.__dir__': ('utils.misc.html#nict.__dir__', 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict.__getattr__': ( 'utils.misc.html#nict.__getattr__',
                                                                                 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict.__init__': ('utils.misc.html#nict.__init__', 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict.__setattr__': ( 'utils.misc.html#nict.__setattr__',
                                                                                 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.nict._to_dict': ('utils.misc.html#nict._to_dict', 'friendlly/utils/misc.py'),
                                      'friendlly.utils.misc.safe_int': ('utils.misc.html#safe_int', 'friendlly/utils/misc.py')},
            'friendlly.utils.notebook': { 'friendlly.utils.notebook.detect_environment': ( 'utils.notebook.html#detect_environment',
                                                                                           'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.extract_vscode_path': ( 'utils.notebook.html#extract_vscode_path',
                                                                                            'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.inject_js': ( 'utils.notebook.html#inject_js',
                                                                                  'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_add_cell': ( 'utils.notebook.html#nbclassic_add_cell',
                                                                                           'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_delete_cell': ( 'utils.notebook.html#nbclassic_delete_cell',
                                                                                              'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_execute_cell': ( 'utils.notebook.html#nbclassic_execute_cell',
                                                                                               'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_get_index': ( 'utils.notebook.html#nbclassic_get_index',
                                                                                            'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_patch_kernel': ( 'utils.notebook.html#nbclassic_patch_kernel',
                                                                                               'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_render_cell': ( 'utils.notebook.html#nbclassic_render_cell',
                                                                                              'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.nbclassic_update_cell': ( 'utils.notebook.html#nbclassic_update_cell',
                                                                                              'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.update_code_next': ( 'utils.notebook.html#update_code_next',
                                                                                         'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.update_code_self': ( 'utils.notebook.html#update_code_self',
                                                                                         'friendlly/utils/notebook.py'),
                                          'friendlly.utils.notebook.vscode_get_cells': ( 'utils.notebook.html#vscode_get_cells',
                                                                                         'friendlly/utils/notebook.py')}}}
